from functools import cached_property
from pathlib import Path
from wizlib.app import WizApp
from wizlib.stream_handler import StreamHandler
from wizlib.config_handler import ConfigHandler
from wizlib.ui_handler import UIHandler

from dyngle.command import DyngleCommand
from dyngle.error import DyngleError
from dyngle.expression import expression
from dyngle.operation import Operation


class DyngleApp(WizApp):

    base = DyngleCommand
    name = 'dyngle'
    handlers = [StreamHandler, ConfigHandler, UIHandler]

    # For possible upstreaming to WizLib, a mechanism to "import" configuration
    # settings from external files.

    @property
    def _imported_configrations(self):
        if not hasattr(self, '__imported_configurations'):
            imports = self.config.get('dyngle-imports')
            confs = []
            if imports:
                for filename in imports:
                    full_filename = Path(filename).expanduser()
                    confs.append(ConfigHandler(full_filename))
            self.__imported_configurations = confs
        return self.__imported_configurations

    def _get_configuration_details(self, type: str):
        label = f'dyngle-{type}'
        details = {}
        for conf in self._imported_configrations:
            if (imported_details := conf.get(label)):
                details |= imported_details
        configured_details = self.config.get(label)
        if configured_details:
            details |= configured_details
        return details

    @cached_property
    def operations(self):
        operations_configs = self._get_configuration_details('operations')
        if not operations_configs:
            raise DyngleError("No operations defined in configuration")
        operations = {}
        for key, config in operations_configs.items():
            if isinstance(config, list):
                operation = Operation({}, config)
            elif isinstance(config, dict):
                local_expr_texts = config.get('expressions') or {}
                local_expressions = _expressions_from_texts(local_expr_texts)
                steps = config.get('steps') or []
                operation = Operation(local_expressions, steps)
            else:
                raise DyngleError(f"Invalid operation configuration for {key}")
            operations[key] = operation
        return operations

    @cached_property
    def global_expressions(self):
        expr_texts = self._get_configuration_details('expressions')
        return _expressions_from_texts(expr_texts)


def _expressions_from_texts(expr_texts):
    if expr_texts:
        return {k: expression(t) for k, t in expr_texts.items()}
    else:
        return {}



from dataclasses import dataclass
from functools import partial
import re

from dyngle.error import DyngleError


PATTERN = re.compile(r'\{\{\s*([^}]+)\s*\}\}')


@dataclass
class Template:

    template: str

    def render(self, data: dict = None, expressions: dict = None) -> str:
        """Render the template with the provided data and expressions.

        Parameters
        ----------
        data : dict
            String data to insert
        expressions : dict
            Functions to call with data

        Returns
        -------
        str
            Template rendered with expression resolution and values inserted.
        """

        data = data if data else {}
        expressions = expressions if expressions else {}
        resolver = partial(self._resolve, live_data=data | expressions)
        return PATTERN.sub(resolver, self.template)

    def _resolve(self, match, *, live_data: dict):
        """Resolve a single name/path from the template. The argument is a
        merge of the raw data and the expressions, either of which are valid
        substitutions."""
        key = match.group(1).strip()
        return self.resolve(key, live_data)

    @staticmethod
    def resolve(key: str, live_data: dict):
        parts = key.split('.')
        current = live_data
        for part in parts:
            if part not in current:
                raise DyngleError(
                    f"Invalid expression or data reference '{key}'")
            current = current[part]
        if callable(current):
            return current(live_data)
        else:
            return current



from dataclasses import dataclass
from functools import cached_property
import re
import shlex
import subprocess

from dyngle.error import DyngleError
from dyngle.template import Template


@dataclass
class Operation:

    local_expressions: dict

    steps: list

    def run(self, data: dict, global_expressions: dict):
        # The data dict is mutable
        steps = self.steps
        expressions = global_expressions | self.local_expressions
        for markup in steps:
            step = Step(markup)
            step.run(data, expressions)


STEP_PATTERN = re.compile(
    r'^\s*(?:([\w.-]+)\s+->\s+)?(.+?)(?:\s+=>\s+([\w.-]+))?\s*$')


def parse_step(markup):
    if match := STEP_PATTERN.match(markup):
        input, command_text, output = match.groups()
        command_template = shlex.split(command_text.strip())
        return input, command_template, output
    else:
        raise DyngleError(f"Invalid step markup {{markup}}")


@dataclass
class Step:

    markup: str

    def __post_init__(self):
        self.input, self.command_template, self.output = \
            parse_step(self.markup)

    def run(self, data, expressions):
        command = [Template(word).render(data, expressions)
                   for word in self.command_template]
        pipes = {}
        if self.output:
            pipes['stdout'] = subprocess.PIPE
        result = subprocess.run(command, text=True, **pipes)
        if result.returncode != 0:
            raise DyngleError(
                f'Step failed with code {result.returncode}: {self.markup}')
        if self.output:
            data[self.output] = result.stdout



import shlex
import subprocess
from wizlib.parser import WizParser
from yaml import safe_load

from dyngle.command import DyngleCommand
from dyngle.expression import expression
from dyngle.template import Template
from dyngle.error import DyngleError


class RunCommand(DyngleCommand):
    """Run a workflow defined in the configuration"""

    name = 'run'

    @classmethod
    def add_args(cls, parser: WizParser):
        super().add_args(parser)
        parser.add_argument('operation', help='Operation name to run')
        parser.add_argument(
            'args', nargs='*', help='Optional operation arguments')

    def handle_vals(self):
        super().handle_vals()

    def _validate_operation_exists(self, operations):
        """Validate that the requested operation exists in configuration"""
        if self.operation not in operations:
            available_operations = ', '.join(operations.keys())
            raise DyngleError(
                f"Operation '{self.operation}' not found. " +
                f"Available operations: {available_operations}")

    @DyngleCommand.wrap
    def execute(self):
        data_string = self.app.stream.text
        data = safe_load(data_string) or {}
        data['args'] = self.args

        operations = self.app.operations
        self._validate_operation_exists(operations)
        operation = operations[self.operation]

        operation.run(data, self.app.global_expressions)

        return f'Operation "{self.operation}" completed successfully'



from typing import Callable

from dyngle.error import DyngleError
from dyngle.safe_path import SafePath

from datetime import datetime as datetime, date, timedelta
import math
import json
import re
import yaml

from dyngle.template import Template


def formatted_datetime(dt: datetime, format_string=None) -> str:
    """Safe datetime formatting using string operations"""
    if format_string is None:
        format_string = "{year:04d}{month:02d}{day:02d}"
    components = {
        'year': dt.year,
        'month': dt.month,
        'day': dt.day,
        'hour': dt.hour,
        'minute': dt.minute,
        'second': dt.second,
        'microsecond': dt.microsecond,
        'weekday': dt.weekday(),  # Monday is 0
    }
    return format_string.format(**components)


GLOBALS = {
    "__builtins__": {
        # Basic data types and conversions
        "int": int,
        "float": float,
        "str": str,
        "bool": bool,
        "list": list,
        "dict": dict,
        "tuple": tuple,
        "set": set,

        # Essential functions
        "len": len,
        "min": min,
        "max": max,
        "sum": sum,
        "abs": abs,
        "round": round,
        "sorted": sorted,
        "reversed": reversed,
        "enumerate": enumerate,
        "zip": zip,
        "range": range,
    },

    # Mathematical operations
    "math": math,

    # Date and time handling
    "datetime": datetime,
    "date": date,
    "timedelta": timedelta,
    "formatted": formatted_datetime,

    # Data parsing and manipulation
    "json": json,
    "yaml": yaml,
    "re": re,

    # Safe Path-like operations (within cwd)
    "Path": SafePath
}


def _evaluate(expression: str, locals: dict) -> str:
    """Evaluate a Python expression with safe globals and user data context.

    Safely evaluates a Python expression string using a restricted set of
    global functions and modules, combined with user-provided data. The
    expression is evaluated in a sandboxed environment that includes basic
    Python built-ins, mathematical operations, date/time handling, and data
    manipulation utilities.

    Parameters
    ----------
    expression : str
        A valid Python expression string to be evaluated.
    data : dict
        Dictionary containing variables and values to be made available during
        expression evaluation. Note that hyphens in keys will be replaced by
        underscores to create valid Python names.

    Returns
    -------
    str
        String representation of the evaluated expression result. If the result
        is a tuple, returns the string representation of the last element.

    Raises
    ------
    DyngleError
        If the expression contains invalid variable names that are not found in
        the provided data dictionary or global context.
    """
    try:
        result = eval(expression, GLOBALS, locals)
    except KeyError:
        raise DyngleError(f"The following expression contains " +
                          f"at least one invalid name: {expression}")
    result = result[-1] if isinstance(result, tuple) else result
    return str(result)


def expression(text: str) -> Callable[[dict], str]:
    def evaluate(data: dict = None) -> str:
        items = data.items() if data else ()
        locals = {k.replace('-', '_'): v for k, v in items}
        def resolve(s): return Template.resolve(s, data)
        return _evaluate(text, locals | {'resolve': resolve})
    return evaluate
